/*
* instruction.cpp
*
* This file contains static helper functions
*
* Copyright 2010 Tobias Wirtl, HBM <tobias.wirtl@hbm.com>
*
* The code contained herein is licensed under the GNU General Public
* License. You may obtain a copy of the GNU General Public License
* Version 2 or later at the following locations:
*
* http://www.opensource.org/licenses/gpl-license.html
* http://www.gnu.org/copyleft/gpl.html
*/

#include "instruction.h"
#include "vector"
#include "helpers.h"
#include <iostream>

Instruction::Instruction(const string &instructionStr, const string &bitMask, const string &options)
{
    unsigned int i;

    this->instructionStr = instructionStr;
    this->bitMask = bitMask;
    for (i=0; i < options.size(); i++)
    {
        switch(options[i])
        {
        case ('r'): this->options |= RELATIVE_JUMP; break;
        case ('a'): this->options |= ABSOLUTE_JUMP; break;
        }
    }
}


//get the bitcode with filled params from a line of code (e.g. MOV 3 6)
string Instruction::getBitcodeOfProgramLine(const string &programLine, int lineNr, map<string, int> *labelList)
{
    unsigned int i1,i2, digitCounter;
    vector<string> *splitLine = new vector<string>();
    string result = this->bitMask;

    Helpers::splitStr(programLine,splitLine,' ');
    if (splitLine->size() >= this->getNumParams() + 1)
    {
        //get bit string of each parameter
        for (i1 = 0; i1 < this->getNumParams(); i1++ )
        {
            string param = (*splitLine)[i1+1];
            digitCounter = 0;
            if (labelList->find(param) != labelList->end())
            {
                if (this->options & RELATIVE_JUMP)
                    param = Helpers::intToStr(labelList->at(param) - lineNr - 1);
                else
                    param = Helpers::intToStr(labelList->at(param));
            }
            string paramBits = this->getBitStr(param, this->getBitLengthParam(i1));
            for (i2 = 0; i2 < this->bitMask.size(); i2++)
            {
                if ((unsigned int)result[i2] == 97+i1)
                    result[i2] = paramBits[digitCounter++];
            }
        }
    }
    return result;
}


//get bitstring of a given parameter string in reg, dec, hex or bin
string Instruction::getBitStr(const string &param, unsigned int length)
{
    int value = 0;
    int negative = 0;
    string result;

    if (param[0] == '-')
        negative = 1;
    //get value of the parameter
    if (param[negative] == 'r')
        value = Helpers::strToInt(param.substr(1+negative,param.length()-(1+negative)));
    else if (param[negative] == 'b')
        value = Helpers::binStrToDecInt(param.substr(1+negative,param.length()-(1+negative)));
    else if (param[negative] == 'h')
        value = Helpers::hexStrToDecInt(param.substr(1+negative,param.length()-(1+negative)));
    else
        value = Helpers::strToInt(param.substr(negative,param.length()-(negative)));

    result = Helpers::intToBinaryStr(value);
    //add leading zeros
    while (result.length() < length)
        result='0'+result;
    if (result.length() > length)
    {
        result = result.substr(result.size()-length,length);
        cout<<"Warning: parameter \""<<param<<"\" was cropped because the value was out of range!\n";
    }

    //build 2nd complement in ascii
    if (negative)
        result = Helpers::build2Complement(result);

    return result;
}


unsigned int Instruction::getNumParams()
{
    char lastParamChar = '0';
    int paramCount = 0;
    for (unsigned int i = 0; i < bitMask.length(); i++)
    {
        if (bitMask[i] != '0' && bitMask[i] != '1' && bitMask[i] != lastParamChar)
        {
            lastParamChar = bitMask[i];
            paramCount++;
        }
    }
    return paramCount;
}


bool Instruction::isAbsoluteJump()
{
    return (options & ABSOLUTE_JUMP);
}


bool Instruction::isRelativeJump()
{
    return (options & RELATIVE_JUMP);
}


//get the bit length of the instructions parameters (zero based)
unsigned int Instruction::getBitLengthParam(int paramNo)
{
    char paramChar = 0x61 + paramNo;
    unsigned int i;
    int counter = 0;
    for (i = 0; i < this->bitMask.length();i++)
        if (bitMask[i] == paramChar)
            counter++;
    return counter;
}

string Instruction::getBitMask()
{
    return this->bitMask;
}

string Instruction::getInstructionStr()
{
    return this->instructionStr;
}

string Instruction::getBitsOfParam(unsigned int paramNo, const string &bitcode)
{
    unsigned int i;
    string result;
    for (i = 0; i < this->bitMask.length();i++)
    {
        if ((unsigned int)this->bitMask[i] == paramNo + 97)
            result.push_back(bitcode[i]);
    }
    return result;
}
